/*****************************************************************
 * Floating Point Co-Processor Interface for the Atari ST line of*
 * Computers.  These routines where tested with a Atari Mega STE *
 * and a 68882 FPCP.                                             *
 *                                                               *
 * These routine should work with any Atari ST with a 68000      *
 * processor and 68881 or 68882 Co-Processor. This code was      *
 * written to be used with the Megamax 'C' Complier. But should  *
 * be usable with any complier that uses the IEEE (754) Double   *
 * Percision Real Numbers. Single, Extend, and BCD real numbers  *
 * are not supported.                                            *
 *                                                               *
 * This code will not work with the more recent processors       *
 * (68020, 68030,...) since the interface is not under software  *
 * only control.                                                 *
 *                                                               *
 * FREEWARE! your are authorized to modify this code to suit     *
 * your needs. However this code either original or modified can *
 * not be sold without the written permission from the author.   *
 * All Copyright notices must remain on the original and any     *
 * Modified versions. All modififed version must also contain    *
 * the name of the modifier and a indicator as to what was       *
 * modified. Take credit for your work.                          *
 *                                                               * 
 * Copyright (C) 1992 By Robert W. Stiles, KD6GEO.               *
 *                                                               *
 * The author is not responsible for use or misuse of this       *
 * product and any ill effects created by them. By using these   *
 * routines you agree to these terms.                            *
 *****************************************************************/

/*****************************************************************
 * Programmers note: When conducting math operations within the  *
 * MCP registers you first must understand the order in which    *
 * the operation are conducted. The following examples will      *
 * illustrate.                                                   *
 *                                                               *
 * Source code equivalents:                                      *
 *                                                               *
 * 'C' source math operation:                                    *
 *                                                               *
 *     double a, b, c;                                           *
 *     c = a * b;                                                *
 *                                                               *
 *  Coprocessor math operation (register to register)            *
 *                                                               *
 *     fmovetoreg(FP0, &a);     Move 'a' value into FP reg 0.    *
 *     fmovetoreg(FP1, &b);     Move 'b' value into FP reg 1.    *
 *     fmul_reg(SFP1|DFP0);     Multiply FP0 with FP1 with the   *
 *                              results being stored in FP0.     *
 *                              The results will always be.      *
 *                              stored in the destination        *
 *                              register.                        *
 *     fmovefromreg(FP0, &c);   Retrieve the results.            *
 *                                                               *
 *     If this ordering was directly convert to a 'C' routine    *
 *     it would appear in this format "c = b * a; ". However the *
 *     order really is not important until subtraction or        *
 *     division is utilized.                                     *
 *                                                               *
 *     All the Trig function work the same as the standard math  *
 *     Lib.                                                      *
 *                                                               *
 *****************************************************************/
 
/*****************************************************************
 * Memory mapped io address of the coprocessor.                  *
 *****************************************************************/

#define  RESPONCE         0xFFFFFA40
#define  CONTROL          0xFFFFFA42
#define  SAVE             0xFFFFFA44
#define  RESTORE          0xFFFFFA46
#define  COMMAND          0xFFFFFA4A
#define  CONDITION        0xFFFFFA4E
#define  OPERAND          0xFFFFFA50
#define  REG_SELECT       0xFFFFFA54
#define  INSTR_ADDR       0xFFFFFA58
#define  OPRAND_ADDR      0xFFFFFA5C

/*****************************************************************
 * Internal Floating Point Register Access Commands.             *
 *****************************************************************/

/*********************************
 * Destination Command Segment   *
 * to be or'ed with the Source   *
 * command segment.              *
 *********************************/
 
#define  DFP0              0x0000
#define  DFP1              0x0080
#define  DFP2              0x0100
#define  DFP3              0x0180
#define  DFP4              0x0200
#define  DFP5              0x0280
#define  DFP6              0x0300
#define  DFP7              0x0380

/*********************************
 * Source Command Segment to be  *
 * or'ed with the Destination    *
 * command segment.              *
 *********************************/

#define  SFP0              0x0000
#define  SFP1              0x0400
#define  SFP2              0x0800
#define  SFP3              0x0c00
#define  SFP4              0x1000
#define  SFP5              0x1400
#define  SFP6              0x1800
#define  SFP7              0x1c00

/********************************
 * The available Math           *
 * CoProcessor registers.       *
 ********************************/
 
#define  FP0               0x00
#define  FP1               0x01
#define  FP2               0x02
#define  FP3               0x03
#define  FP4               0x04
#define  FP5               0x05
#define  FP6               0x06
#define  FP7               0x07

/*****************************************************************
 * Math functions.                                               *
 *****************************************************************/
#define  FMOVETOFPN       0x00      /* Move to FPn Register           */
#define  FINT             0x01      /* Integer part of Source to FPn. */
#define  FSINH            0x02      /* Hyperbolic Sine.               */
#define  FINTRZ           0x03      /* Integer part rounded to zero.  */
#define  FSQRT            0x04      /* Square Root.                   */
#define  FLOGNP1          0x06      /* Log e (x + 1).                 */
#define  FETOXM1          0x08      /* e (source) - 1.                */
#define  FTANH            0x09      /* Hyperbolic Tangent.            */
#define  FATAN            0x0A      /* Arc Tangent.                   */
#define  FASIN            0x0C      /* Arc Sine.                      */
#define  FATANH           0x0D      /* Hyperbolic Arc Tangent.        */
#define  FSIN             0x0E      /* Sine.                          */
#define  FTAN             0x0F      /* Tangent.                       */
#define  FETOX            0x10      /* e (source).                    */
#define  FTWOTOX          0x11      /* 2 to the power of (source).    */
#define  FTENTOX          0x12      /* 10 to the power of (source).   */
#define  FLOGN            0x14      /* Log e of (source).             */
#define  FLOG10           0x15      /* Log 10 of (source).            */
#define  FLOG2            0x16      /* Log 2 of (source).             */
#define  FABS             0x18      /* Absolute Value of (source).    */
#define  FCOSH            0x19      /* Hyperbolic Cosine.             */
#define  FNEG             0x1A      /* Negate value.                  */
#define  FACOS            0x1C      /* Arc Cosine.                    */
#define  FCOS             0x1D      /* Cosine.                        */
#define  FGETEXP          0x1E      /* Get Exponent of value.         */
#define  FGETMAN          0x1F      /* Get Mantissa of value.         */
#define  FDIV             0x20      /* Divide.                        */
#define  FMOD             0x21      /* Modulo Remainder.              */
#define  FADD             0x22      /* Add.                           */
#define  FMUL             0x23      /* Multiply.                      */
#define  FREM             0x25      /* Remainder of Value.            */
#define  FSCALE           0x26      /* Scale Exponent.                */
#define  FSUB             0x28      /* Subtract.                      */
#define  FCMP             0x38      /* Compare Values.                */
#define  FTST             0x3A      /* Test Value.                    */

/**********************************************************************
 * Within the CP there lay constant values which can be loaded        *
 * directly to any register.  The follow is a list of them for use.   *
 **********************************************************************/
 
#define  PI               0x00
#define  LOG10TWO         0x0B
#define  E                0x0C
#define  LOG2E            0x0D
#define  LOG10E           0x0E
#define  ZERO             0x0F

extern int fadd_reg(); 
extern int fsub_reg();
extern int fmul_reg();
extern int fdiv_reg();
extern int fcmp_reg();
extern int fmod_reg();
extern int frem_reg();

extern int fmovetoreg();
extern int fmovefromreg();
extern int fmove_rr();
extern int fmovecr();
extern int fp_regtoreg();
extern int wake_fpcp();
extern int do_trig();

extern double fasin();
extern double fsin();
extern double fsinh();
extern double fcos();
extern double fcosh();
extern double facos();
extern double ftan();
extern double fatan();
extern double ftanh();
extern double fatanh();
extern double ffabs();
extern double fneg();

extern double _fpreg0;

unsigned int fpmc_ea_reg[] = {  
   0x5400, 0x5480, 0x5500, 0x5580, 0x5600, 0x5680, 0x5700, 0x5780 };

unsigned int fpmc_reg_ea[] = { 
   0x7400, 0x7480, 0x7500, 0x7580, 0x7600, 0x7680, 0x7700, 0x7780 };
   
unsigned int fp_rdst[] = { 
   0x0000, 0x0080, 0x0100, 0x0180, 0x0200, 0x0280, 0x0300, 0x0380 };

asm
{                     
;*****************************************************************
;* Add the contents of one register to another.                  *
;* Return 0 if OKay,  Responce register code if not.             *
;* int fadd_reg(int src_register or'ed dst_register);            *
;*****************************************************************
fadd_reg:     move.w  4(A7),D1            ;* Registers Used.
              move.w  #FADD,D2            ;* Math Operation
              jsr     fp_regtoreg
              rts
              
;*****************************************************************
;* Subtract the contents of one register to another.             *
;* Return 0 if OKay,  Responce register code if not.             *
;* int fsub_reg(int src_register or'ed dst_register);            *
;*****************************************************************
fsub_reg:     move.w  4(A7),D1            ;* Registers Used.
              move.w  #FSUB,D2            ;* Math Operation.
              jsr     fp_regtoreg
              rts

;*****************************************************************
;* Multiply the contents of one register to another.             *
;* Return 0 if OKay,  Responce register code if not.             *
;* int fmul_reg(int src_register or'ed dst_register);            *
;*****************************************************************
fmul_reg:     move.w  4(A7),D1            ;* Registers Used.
              move.w  #FMUL,D2            ;* Math Operation.
              jsr     fp_regtoreg
              rts

;*****************************************************************
;* Divide the contents of one register to another.               *
;* Return 0 if OKay,  Responce register code if not.             *
;* int fdiv_reg(int src_register or'ed dst_register);            *
;*****************************************************************
fdiv_reg:     move.w  4(A7),D1            ;* Registers Used.
              move.w  #FDIV,D2            ;* Math Operation
              jsr     fp_regtoreg
              rts

;*****************************************************************
;* Modulo Remainder the contents of one register to another.     *
;* Return 0 if OKay,  Responce register code if not.             *
;* int fmod_reg(int src_register or'ed dst_register);            *
;*****************************************************************
fmod_reg:     move.w  4(A7),D1            ;* Registers Used.
              move.w  #FMOD,D2            ;* Math Operation
              jsr     fp_regtoreg
              rts

;*****************************************************************
;* IEEE Remainder the contents of one register to another.       *
;* Return 0 if OKay,  Responce register code if not.             *
;* int frem_reg(int src_register or'ed dst_register);            *
;*****************************************************************
frem_reg:     move.w  4(A7),D1            ;* Source/Dest Register.
              move.w  #FREM,D2            ;* Math Operation
              jsr     fp_regtoreg
              rts

;*****************************************************************
;* Compare the contents of one register to another.              *
;* Returned value of 0 = equal, 1 = greater, -1 = less than.     *
;* example: SFP0 = 1.0, DFP1 = 2.0, fcmp_reg(SFP0|DFP1) ==  1    *
;*          SFP0 = 2.0, DFP1 = 1.0. fcmp_reg(SFP0|DFP1) == -1    *
;*          SFP0 = 1.0, DFP1 = 1.0. fcmp_reg(SFP0|DFP1) ==  0    *
;* int fcmp_reg(int src_register or'ed dst_register);            *
;*****************************************************************
fcmp_reg:     move.w  4(A7),D1            ;* Source/Dest Register.
              move.w  #FCMP,D2            ;* Compare Operation.
              jsr     fp_regtoreg
next_op:
              move.w  #0xA800, COMMAND    ;* Send command to read the
                                          ;* status register.
              move.w  #0x7f,D1            ;* Number of tries. 

loopz:        move.w  RESPONCE,D0
              cmpi.w  #0xB104,D0
              beq     L4                  ;* Go read FPSR.
              dbf     D1,loopz            ;* Recheck responce register.
              
              move.w  D0,-(A7)            ;* Save responce.
              jsr     wake_fpcp           ;* Wake up MCP.
              move.w  (A7)+,D0            ;* Restore register value.
              rts                         ;* Bad news boys!
              
L4:           move.l  OPERAND,D0          ;* Status Register Contents.
              moveq   #27,D1              ;* Prepare for bit check.
              btst    D1,D0               ;* Check the negative bit.
              beq     L5                  ;* Not less then.
              moveq   #-1,D0              ;* It's Less then.
              rts                         ;* Return to the caller.
                         
L5:           subq.w  #1,D1               ;* Prepare for next bit.
              btst    D1,D0               ;* Equal? (Zero bit).
              beq    L6                   ;* Not equal to.
              moveq   #0,D0               ;* It's equal to.
              rts                         ;* return to caller.
              
L6:           moveq   #1,D0               ;* So it must be greater than.
              rts                         ;* Return to caller.
              
;*****************************************************************
;* Sine the contents of FP0 register.                            *
;* double fsin(double radians);                                  *
;*****************************************************************
fsin:         move.l  4(A7),D0            ;* The data.
              move.l  8(A7),D1            
              move.w  #FSIN,D2            ;* Math Operation
              jsr     do_trig
              rts

;*****************************************************************
;* Arc Sine the contents of FP0 register.                        *
;* double fasin(double radians);                                 *
;*****************************************************************
fasin:        move.l  4(A7),D0           ;* The data.
              move.l  8(A7),D1            
              move.w  #FASIN,D2          ;* Math Operation
              jsr     do_trig
              rts

;*****************************************************************
;* Hyperbolic Sine the contents of FP0 register.                 *
;* double fsinh(double radians);                                 *
;*****************************************************************
fsinh:        move.l  4(A7),D0            ;* The data.
              move.l  8(A7),D1            
              move.w  #FSINH,D2           ;* Math Operation
              jsr     do_trig
              rts

;*****************************************************************
;* Tangent the contents of FP0 register.                         *
;* double ftan(double radians);                                  *
;*****************************************************************
ftan:         move.l  4(A7),D0          ;* The data.
              move.l  8(A7),D1            
              move.w  #FTAN,D2          ;* Math Operation
              jsr     do_trig
              rts

;*****************************************************************
;* Arc Tangent the contents of FP0 register.                     *
;* double fatan(double radians);                                 *
;*****************************************************************
fatan:        move.l  4(A7),D0            ;* The data.
              move.l  8(A7),D1            
              move.w  #FATAN,D2           ;* Math Operation
              jsr     do_trig
              rts


;*****************************************************************
;* Hyperbolic Tangent the contents of FP0 register.              *
;* double ftanh(double radians);                                 *
;*****************************************************************
ftanh:        move.l  4(A7),D0            ;* The data.
              move.l  8(A7),D1            
              move.w  #FTANH,D2           ;* Math Operation
              jsr     do_trig
              rts

;*****************************************************************
;* Arc Hyperbolic Tangent the contents of FP0 register.          *
;* double fatanh(double radians);                                *
;*****************************************************************
fatanh:       move.l  4(A7),D0           ;* The data.
              move.l  8(A7),D1            
              move.w  #FATANH,D2         ;* Math Operation
              jsr     do_trig
              rts


;*****************************************************************
;* Cosine the contents of FP0 register.                          *
;* double fcos(double radians);                                  *
;*****************************************************************
fcos:         move.l  4(A7),D0          ;* The data.
              move.l  8(A7),D1            
              move.w  #FCOS,D2          ;* Math Operation
              jsr     do_trig
              rts

;*****************************************************************
;* Arc Cosine the contents of FP0 register.                      *
;* double facos(double radians);                                 *
;*****************************************************************
facos:        move.l  4(A7),D0          ;* The data.
              move.l  8(A7),D1            
              move.w  #FACOS,D2         ;* Math Operation
              jsr     do_trig
              rts

;*****************************************************************
;* Hyperbolic Cosine the contents of FP0 register.               *
;* double fcosh(double radians);                                 *
;*****************************************************************
fcosh:        move.l  4(A7),D0          ;* The data.
              move.l  8(A7),D1            
              move.w  #FCOSH,D2         ;* Math Operation
              jsr     do_trig
              rts

;*****************************************************************
;* Floating point Absolute the contents of FP0 register.         *
;* double ffabs(double number);                                  *
;*****************************************************************
ffabs:        move.l  4(A7),D0          ;* The data.
              move.l  8(A7),D1            
              move.w  #FABS,D2          ;* Math Operation
              jsr     do_trig
              rts

;*****************************************************************
;* Negate the contents of FP0 register.                          *
;* double fneg(double number);                                   *
;*****************************************************************
fneg:         move.l  4(A7),D0          ;* The data.
              move.l  8(A7),D1            
              move.w  #FNEG,D2          ;* Math Operation
              jsr     do_trig
              rts

;*****************************************************************
;* Move the contents of one register to another.                 *
;* Return 0 if OKay,  Responce register code if not.             *
;* int fmove_rr(int src_register or'ed dst_register);            *
;*****************************************************************
fmove_rr:     move.w  4(A7),D1            ;* Source Register.
              move.w  #FMOVETOFPN,D2      ;* Fmove Operation.
              jsr     fp_regtoreg
              rts

;*****************************************************************
;* Move the constant value to a register from MCP ROM.           *
;* Return 0 if OKay,  Responce register code if not.             *
;* int fmovecr(int register, int constant);                      *
;*****************************************************************
fmovecr:      jsr     wake_fpcp

              move.w  4(A7),D0            ;* Destination Register.
              move.w  6(A7),D1            ;* Rom storage loc. number.
              lea     fp_rdst,A0          ;* Destination register table.
              andi.w  #0x07,D0            ;* Keep number small.
              move.w  0(A0,D0.W),D0       ;* The register to load.
              or.w    D1,D0               ;* Constant selected.
              ori.w   #0x5E00,D0          ;* The command to send.
             
              move.w  D0,COMMAND          ;* Start the sequence.
              move.w  #0x1000,D1          ;* Number of tries.
 loopa:       move.w  RESPONCE,D0         ;* Read responce register.
              cmpi.w  #0x802,D0           ;* At idle state?
              beq     all_okay            ;* if so, command completed.
              dbf     D1,loopa            ;* if not try again.
              move.w  D0,-(A7)
              jsr     wake_fpcp           ;* wake up the MCP.
              move.w  (A7)+,D0
              rts                         ;* Bad news boys.
all_okay:     moveq   #0,D0               ;* It must have worked.
              rts                         ;* Return to the caller.

;*****************************************************************
;* Handles the process of doing the trig calculation in the      *
;* registers to the outside world.                               *
;*****************************************************************
do_trig:
    ori.w    #0x5400,D2                   ;* Register 0         
                                          ;* Or'ed with math function.
    move.w   D2,COMMAND                   ;* Send the command.
loopb:
    move.w   RESPONCE,D2                  ;* Read responce register.
    andi.w   #0x0fff,D2                   ;* Mask out unwanted bits.
    cmpi.w   #0x0608,D2                   ;* Ready for data?
    bne      loopb                        ;* If not recheck.
    move.l   D0,OPERAND                   ;* send the data first 32 bits.
    move.l   D1,OPERAND                   ;* second 32 bits.
loopd:
    move.w   RESPONCE,D2                  ;* Read responce register.
    cmpi.w   #0x0802,D2                   ;* Operation Complete? (idle).
    bne      loopd                        ;* If not recheck until so.
    move.w   #0x7400,COMMAND              ;* Command to read double prec
                                          ;* from the operand registers.
loope:
    move.w   RESPONCE,D2                  ;* Read the resopnce register.
    andi.w   #0x0fff,D2                   ;* Mask out unwanted data.
    cmpi.w   #0x0208,D2                   ;* Ready to transfer data?
    bne      loope                        ;* If not recheck responce reg.
    move.l   OPERAND,_fpreg0              ;* Move the data to.
    move.l   OPERAND,_fpreg0+4            ;* Megamax Laser 'C' storeage loc.
    rts                                   ;* Return to the caller.
    
;*****************************************************************
;* Handles the process of doing the calculation in the           *
;* registers.                                                    *
;*****************************************************************
fp_regtoreg: 
              or.w    D2,D1               ;* Build the command.
                            
              move.w  D1,COMMAND          ;* Tell the copressor to do it.
              move.w  #0x7f,D0            ;* Number of tries.        

loop1:        move.w  RESPONCE,D2         ;* Check Status of coprocessor.
              cmpi.w  #0x0802,D2          ;* Match? (idle)
              beq     process_ok          ;* If so Branch.
              dbf     D0,loop1            ;* If not try again.

              move.w  D2,-(A7)
              jsr     wake_fpcp           ;* wake up the MCP.
              move.w  (A7)+,D0
              rts                         ;* Bad news boys.
process_ok:
              moveq   #0,D0               ;* Must of worked.
              rts                         ;* Return to caller.
                             
              
;*****************************************************************
;* Perform indirect to FP register move. Double pointer and      *
;* register on the stack.                                        *
;* Return 1 if Okay, 0 if not.                                   *
;* int fmovetoreg(int register, double *value);                  *
;*****************************************************************

fmovetoreg:  
             move.w  4(A7),D0             ;* Register to load.
             move.l  6(A7),A0             ;* Value to move.

             andi.w  #0x7,D0              ;* Mask so dont overrun data.
             lsl.w   #1,D0                ;* Shift for word size.
             lea     fpmc_ea_reg,A1       ;* Address of commands.
             move.w  0(A1,D0.W),D0        ;* The command to send.
            
             move.w  #0x7f,D1             ;* Number of tries before fail.
             move.w  D0,COMMAND           ;* Send command to fmove to FPn
             
again1:      move.w  RESPONCE,D0          ;* Check FPCP Responce.
             andi.w  #0x0FFF,D0           ;* Don't care about upper bits.          
             cmpi.w  #0x0608,D0           ;* One of two responces.
             beq     send_fmove1          ;* Send the data.
             cmpi.w  #0x0208,D0           ;* The other responce.
             beq     send_fmove1          ;* Send the data.
             dbf     D1,again1            ;* If not read try again.

             move.w  D0,-(A7)
             jsr     wake_fpcp            ;* wake up the MCP.
             move.w  (A7)+,D0
             rts                          ;* Bad news boys.
              
send_fmove1: move.l  0(A0),OPERAND        ;* Send the first 32 bits.
             move.l  4(A0),OPERAND        ;* Send the second 32 bits.
             moveq   #1,D0                ;* Okay, return a one status.
             rts                          ;* Return to the caller.

;*****************************************************************
;* Perform FP register move to indirect A0. Double pointer       *
;* and register on the stack.                                    *
;* Return 1 if Okay, 0 if not.                                   *
;* int fmovefromreg(int register, double *value);                *
;*****************************************************************

fmovefromreg:  
             move.w  4(A7),D0             ;* Register to read.
             move.l  6(A7),A0             ;* Move to value to address.

             andi.w  #0x7,D0              ;* Mask so dont overrun data.
             lsl.w   #1,D0                ;* Shift for word size.
             lea     fpmc_reg_ea,A1       ;* Address of commands.
             move.w  0(A1,D0.W),D0        ;* The command to send.
            
             move.w  #0x7f,D1             ;* Number of tries before fail.
             move.w  D0,COMMAND           ;* Send command to fmove to FPn
             
again2:      move.w  RESPONCE,D0          ;* Check FPCP Responce.
             andi.w  #0x0FFF,D0           ;* Don't care about upper bits.          
             cmpi.w  #0x0208,D0           ;* One of two responces.
             beq     send_fmove2          ;* Send the data.
             cmpi.w  #0x0608,D0           ;* The other responce.
             beq     send_fmove2          ;* Send the data.
             dbf     D1,again2            ;* If not read try again.
             move.w  D0,-(A7)
             jsr     wake_fpcp            ;* wake up the MCP.
             move.w  (A7)+,D0
             rts                          ;* Bad news boys.
             
send_fmove2: move.l  OPERAND,0(A0)        ;* Move the first 32 bits.
             move.l  OPERAND,4(A0)        ;* Move the second 32 bits.
             moveq   #1,D0                ;* Okay, return a one status.
             rts                          ;* Return to the caller.
             
;*****************************************************************
;* Make sure the coprocessor is alive to take commands.          *
;*****************************************************************

wake_fpcp:  move.w   #0,CONTROL           ;* Wake up!
            move.w   #0x7f,D1             ;* Number of tries.
wake1:      move.w   RESPONCE,D0          ;* Check Responce.
            cmpi.w   #0x0802,D0           ;* In a idle State?
            beq      done                 ;* yes.
            dbf      D1,wake1             ;* if not try again.
done:            
            rts                           ;* If okay return.
            
}
